@@ -31,10 +31,6 @@
 
 #include "config-plasma.h"
 
-#if !PLASMA_NO_KIO
-#include <kio/job.h>
-#endif
-
 #include "applet.h"
 #include "containment.h"
 #include "containmentactions.h"
@@ -62,7 +58,6 @@ public:
     }
 
     static QSet<QString> knownCategories();
-    static QString parentAppConstraint(const QString &parentApp = QString());
 
     static QSet<QString> s_customCategories;
     QHash<QString, QPointer<PackageStructure> > structures;
@@ -74,15 +69,23 @@ public:
     static QString s_servicesPluginDir;
     static QString s_containmentActionsPluginDir;
 
-    // We only use this cache during start of the process to speed up many consecutive calls
-    // After that, we're too afraid to produce race conditions and it's not that time-critical anyway
-    // the 20 seconds here means that the cache is only used within 20sec during startup, after that,
-    // complexity goes up and we'd have to update the cache in order to avoid subtle bugs
-    // just not using the cache is way easier then, since it doesn't make *that* much of a difference,
-    // anyway
-    int maxCacheAge = 20;
-    qint64 pluginCacheAge = 0;
-    QHash<QString, QVector<KPluginMetaData>> pluginCache;
+    class Cache {
+        // We only use this cache during start of the process to speed up many consecutive calls
+        // After that, we're too afraid to produce race conditions and it's not that time-critical anyway
+        // the 20 seconds here means that the cache is only used within 20sec during startup, after that,
+        // complexity goes up and we'd have to update the cache in order to avoid subtle bugs
+        // just not using the cache is way easier then, since it doesn't make *that* much of a difference,
+        // anyway
+        int maxCacheAge = 20;
+        qint64 pluginCacheAge = 0;
+        QHash<QString, QVector<KPluginMetaData>> plugins;
+
+    public:
+        QVector<KPluginMetaData> findPluginsById(const QString& name, const QStringList &dirs);
+    };
+    Cache plasmoidCache;
+    Cache dataengineCache;
+    Cache containmentactionCache;
 };
 
 QSet<QString> PluginLoaderPrivate::s_customCategories;
@@ -95,7 +98,7 @@ QString PluginLoaderPrivate::s_containmentActionsPluginDir = QStringLiteral("pla
 
 QSet<QString> PluginLoaderPrivate::knownCategories()
 {
-    // this is to trick the tranlsation tools into making the correct
+    // this is to trick the translation tools into making the correct
     // strings for translation
     QSet<QString> categories = s_customCategories;
     categories << QStringLiteral(I18N_NOOP("Accessibility")).toLower()
@@ -123,21 +126,6 @@ QSet<QString> PluginLoaderPrivate::knownCategories()
     return categories;
 }
 
-QString PluginLoaderPrivate::parentAppConstraint(const QString &parentApp)
-{
-    if (parentApp.isEmpty()) {
-        QCoreApplication *app = QCoreApplication::instance();
-        if (!app) {
-            return QString();
-        }
-
-        return QStringLiteral("((not exist [X-KDE-ParentApp] or [X-KDE-ParentApp] == '') or [X-KDE-ParentApp] == '%1')")
-               .arg(app->applicationName());
-    }
-
-    return QStringLiteral("[X-KDE-ParentApp] == '%1'").arg(parentApp);
-}
-
 PluginLoader::PluginLoader()
     : d(new PluginLoaderPrivate)
 {
@@ -191,52 +179,19 @@ Applet *PluginLoader::loadApplet(const QString &name, uint appletId, const QVari
         appletId = ++AppletPrivate::s_maxAppletId;
     }
 
-    const qint64 now = qRound64(QDateTime::currentMSecsSinceEpoch() / 1000.0);
-    bool useRuntimeCache = true;
-    if (now - d->pluginCacheAge > d->maxCacheAge && d->pluginCacheAge != 0) {
-        // cache is old and we're not within a few seconds of startup anymore
-        useRuntimeCache = false;
-        d->pluginCache.clear();
-    }
-
-    if (d->pluginCacheAge == 0) {
-        // Find all the plugins now, but only once
-        d->pluginCacheAge = now;
-
-        auto insertIntoCache = [this](const QString &pluginPath) {
-            KPluginMetaData metadata(pluginPath);
-            if (!metadata.isValid()) {
-                return;
-            }
-
-            d->pluginCache[metadata.pluginId()].append(metadata);
-        };
-
-        KPluginLoader::forEachPlugin(PluginLoaderPrivate::s_plasmoidsPluginDir, insertIntoCache);
-        // COMPAT CODE for applets installed into the toplevel plugins dir by mistake.
-        KPluginLoader::forEachPlugin(QString(), insertIntoCache);
-    }
-
-    //if name wasn't a path, pluginName == name
-    const QString pluginName = name.section(QLatin1Char('/'), -1);
-
-    QVector<KPluginMetaData> plugins;
-
-    if (useRuntimeCache) {
-        auto it = d->pluginCache.constFind(pluginName);
-        if (it != d->pluginCache.constEnd()) {
-            plugins = *it;
-        }
-    } else {
-        plugins = KPluginLoader::findPluginsById(PluginLoaderPrivate::s_plasmoidsPluginDir, pluginName);
-        // COMPAT CODE for applets installed into the toplevel plugins dir by mistake.
-        if (plugins.isEmpty()) {
-            plugins = KPluginLoader::findPluginsById(QString(), pluginName);
-        }
-    }
+    // Need to pass the empty directory because it's where plasmoids used to be
+    auto plugins = d->plasmoidCache.findPluginsById(name, { PluginLoaderPrivate::s_plasmoidsPluginDir, {} });
 
     const KPackage::Package p = KPackage::PackageLoader::self()->loadPackage(QStringLiteral("Plasma/Applet"), name);
 
+    // If the applet is using another applet package, search for the plugin of the other applet
+    if (plugins.isEmpty()) {
+        const QString parentPlugin = p.metadata().value(QStringLiteral("X-Plasma-RootPath"));
+        if (!parentPlugin.isEmpty()) {
+            plugins = d->plasmoidCache.findPluginsById(parentPlugin, { PluginLoaderPrivate::s_plasmoidsPluginDir, {} });
+        }
+    }
+
     if (!plugins.isEmpty()) {
         KPluginLoader loader(plugins.first().fileName());
         if (!isPluginVersionCompatible(loader)) {
@@ -260,7 +215,7 @@ Applet *PluginLoader::loadApplet(const QString &name, uint appletId, const QVari
         QVariantList allArgs;
         allArgs << QVariant::fromValue(p) << p.metadata().fileName() << appletId << args;
 
-        if (p.metadata().serviceTypes().contains(QStringLiteral("Plasma/Containment"))) {
+        if (p.metadata().serviceTypes().contains(QLatin1String("Plasma/Containment"))) {
             applet = new Containment(nullptr, allArgs);
         } else {
             applet = new Applet(nullptr, allArgs);
@@ -283,19 +238,12 @@ DataEngine *PluginLoader::loadDataEngine(const QString &name)
     }
 
     // Look for C++ plugins first
-    auto filter = [&name](const KPluginMetaData &md) -> bool
-    {
-        return md.pluginId() == name;
-    };
-    QVector<KPluginMetaData> plugins = KPluginLoader::findPlugins(PluginLoaderPrivate::s_dataEnginePluginDir, filter);
-
+    const QVector<KPluginMetaData> plugins = d->dataengineCache.findPluginsById(name, {PluginLoaderPrivate::s_dataEnginePluginDir});
     if (!plugins.isEmpty()) {
-        KPluginLoader loader(plugins.first().fileName());
+        KPluginLoader loader(plugins.constFirst().fileName());
         const QVariantList argsWithMetaData = QVariantList() << loader.metaData().toVariantMap();
         KPluginFactory *factory = loader.factory();
-        if (factory) {
-            engine = factory->create<Plasma::DataEngine>(nullptr, argsWithMetaData);
-        }
+        return factory ? factory->create<Plasma::DataEngine>(nullptr, argsWithMetaData) : nullptr;
     }
     if (engine) {
         return engine;
@@ -426,20 +374,13 @@ ContainmentActions *PluginLoader::loadContainmentActions(Containment *parent, co
         return actions;
     }
 
-
-    // Look for C++ plugins first
-    auto filter = [&name](const KPluginMetaData &md) -> bool
-    {
-        return md.pluginId() == name;
-    };
-    QVector<KPluginMetaData> plugins = KPluginLoader::findPlugins(PluginLoaderPrivate::s_containmentActionsPluginDir, filter);
+    const QVector<KPluginMetaData> plugins = d->containmentactionCache.findPluginsById(name, {PluginLoaderPrivate::s_containmentActionsPluginDir});
 
     if (!plugins.isEmpty()) {
         KPluginLoader loader(plugins.first().fileName());
-        const QVariantList argsWithMetaData = QVariantList() << loader.metaData().toVariantMap();
         KPluginFactory *factory = loader.factory();
         if (factory) {
-            actions = factory->create<Plasma::ContainmentActions>(nullptr, argsWithMetaData);
+            actions = factory->create<Plasma::ContainmentActions>(nullptr, {QVariant::fromValue(plugins.first())});
         }
     }
     if (actions) {
@@ -479,7 +420,6 @@ ContainmentActions *PluginLoader::loadContainmentActions(Containment *parent, co
     return actions;
 }
 
-#ifndef PLASMA_NO_DEPRECATED
 Package PluginLoader::loadPackage(const QString &packageFormat, const QString &specialization)
 {
     if (!d->isDefaultLoader) {
@@ -526,7 +466,6 @@ Package PluginLoader::loadPackage(const QString &packageFormat, const QString &s
 
     return Package();
 }
-#endif
 
 QList<KPluginMetaData> PluginLoader::listAppletMetaData(const QString &category, const QString &parentApp)
 {
@@ -539,7 +478,7 @@ QList<KPluginMetaData> PluginLoader::listAppletMetaData(const QString &category,
         filter = [excluded, parentApp](const KPluginMetaData &md) -> bool
         {
             const QString pa = md.value(QStringLiteral("X-KDE-ParentApp"));
-            return (pa.isEmpty() || pa == parentApp) && !excluded.contains(md.category());
+            return (parentApp.isEmpty() || pa == parentApp) && !excluded.contains(md.category());
         };
     } else { //specific category (this could be an excluded one - is that bad?)
 
@@ -547,9 +486,9 @@ QList<KPluginMetaData> PluginLoader::listAppletMetaData(const QString &category,
         {
             const QString pa = md.value(QStringLiteral("X-KDE-ParentApp"));
             if (category == QLatin1String("Miscellaneous")) {
-                return (pa.isEmpty() || pa == parentApp) && (md.category() == category || md.category().isEmpty());
+                return (parentApp.isEmpty() || pa == parentApp) && (md.category() == category || md.category().isEmpty());
             } else {
-                return (pa.isEmpty() || pa == parentApp) && md.category() == category;
+                return (parentApp.isEmpty() || pa == parentApp) && md.category() == category;
             }
         };
     }
@@ -567,7 +506,7 @@ KPluginInfo::List PluginLoader::listAppletInfo(const QString &category, const QS
     const auto plugins = listAppletMetaData(category, parentApp);
 
     //NOTE: it still produces kplugininfos from KServices because some user code expects
-    //info.sevice() to be valid and would crash ohtherwise
+    //info.service() to be valid and would crash otherwise
     foreach (auto& md, plugins) {
         auto pi = md.metaDataFileName().endsWith(QLatin1String(".json")) ? KPluginInfo(md) : KPluginInfo(KService::serviceByStorageId(md.metaDataFileName()));
         if (!pi.isValid()) {
@@ -604,7 +543,7 @@ QList<KPluginMetaData> PluginLoader::listAppletMetaDataForUrl(const QUrl &url)
     auto filter = [&parentApp](const KPluginMetaData &md) -> bool
     {
         const QString pa = md.value(QStringLiteral("X-KDE-ParentApp"));
-        return (pa.isEmpty() || pa == parentApp) && !KPluginMetaData::readStringList(md.rawData(), QStringLiteral("X-Plasma-DropUrlPatterns")).isEmpty();
+        return (parentApp.isEmpty() || pa == parentApp) && !KPluginMetaData::readStringList(md.rawData(), QStringLiteral("X-Plasma-DropUrlPatterns")).isEmpty();
     };
     const QList<KPluginMetaData> allApplets = KPackage::PackageLoader::self()->findPackages(QStringLiteral("Plasma/Applet"), QString(), filter);
 
@@ -638,7 +577,7 @@ QStringList PluginLoader::listAppletCategories(const QString &parentApp, bool vi
     auto filter = [&parentApp, &excluded, visibleOnly](const KPluginMetaData &md) -> bool
     {
         const QString pa = md.value(QStringLiteral("X-KDE-ParentApp"));
-        return (pa.isEmpty() || pa == parentApp)
+        return (parentApp.isEmpty() || pa == parentApp)
             && (excluded.isEmpty() || excluded.contains(md.value(QStringLiteral("X-KDE-PluginInfo-Category"))))
             && (!visibleOnly || !md.isHidden());
     };
@@ -694,14 +633,12 @@ KPluginInfo::List PluginLoader::listContainmentsOfType(const QString &type,
         const QString &parentApp)
 {
     KConfigGroup group(KSharedConfig::openConfig(), "General");
-    const QStringList excluded = group.readEntry("ExcludeCategories", QStringList());
     auto filter = [&type, &category, &parentApp](const KPluginMetaData &md) -> bool
     {
-        if (!md.serviceTypes().contains(QStringLiteral("Plasma/Containment"))) {
+        if (!md.serviceTypes().contains(QLatin1String("Plasma/Containment"))) {
             return false;
         }
-        const QString pa = md.value(QStringLiteral("X-KDE-ParentApp"));
-        if (!pa.isEmpty() && pa != parentApp) {
+        if (!parentApp.isEmpty() && md.value(QStringLiteral("X-KDE-ParentApp")) != parentApp) {
             return false;
         }
 
@@ -754,9 +691,7 @@ KPluginInfo::List PluginLoader::listDataEngineInfo(const QString &parentApp)
     }
 
     QString constraint;
-    if (parentApp.isEmpty()) {
-        constraint = QStringLiteral("not exist [X-KDE-ParentApp]");
-    } else {
+    if (!parentApp.isEmpty()) {
         constraint = QLatin1String("[X-KDE-ParentApp] == '") + parentApp + QLatin1Char('\'');
     }
 
@@ -773,9 +708,7 @@ KPluginInfo::List PluginLoader::listContainmentActionsInfo(const QString &parent
     }
 
     QString constraint;
-    if (parentApp.isEmpty()) {
-        constraint = QStringLiteral("not exist [X-KDE-ParentApp]");
-    } else {
+    if (!parentApp.isEmpty()) {
         constraint = QLatin1String("[X-KDE-ParentApp] == '") + parentApp + QLatin1Char('\'');
     }
 
@@ -828,14 +761,12 @@ Service *PluginLoader::internalLoadService(const QString &name, const QVariantLi
 }
 
 
-#ifndef PLASMA_NO_DEPRECATED
 Package PluginLoader::internalLoadPackage(const QString &name, const QString &specialization)
 {
     Q_UNUSED(name);
     Q_UNUSED(specialization);
     return Package();
 }
-#endif
 
 KPluginInfo::List PluginLoader::internalAppletInfo(const QString &category) const
 {
@@ -909,16 +840,62 @@ bool PluginLoader::isPluginVersionCompatible(KPluginLoader &loader)
     const quint32 maxVersion = PLASMA_MAKE_VERSION(PLASMA_VERSION_MAJOR, PLASMA_VERSION_MINOR, 60);
 
     if (version < minVersion || version > maxVersion) {
-#ifndef NDEBUG
-        qCDebug(LOG_PLASMA) << loader.fileName() << ": this plugin is compiled against incompatible Plasma version" << version
+        qCWarning(LOG_PLASMA) << loader.fileName() << ": this plugin is compiled against incompatible Plasma version" << version
                 << "This build is compatible with" << PLASMA_VERSION_MAJOR << ".0.0 (" << minVersion
                 << ") to" << PLASMA_VERSION_STRING << "(" << maxVersion << ")";
-#endif
         return false;
     }
 
     return true;
 }
 
-} // Plasma Namespace
+QVector<KPluginMetaData> PluginLoaderPrivate::Cache::findPluginsById(const QString& name, const QStringList &dirs)
+{
+    const qint64 now = qRound64(QDateTime::currentMSecsSinceEpoch() / 1000.0);
+    bool useRuntimeCache = true;
 
+    if (pluginCacheAge == 0) {
+        // Find all the plugins now, but only once
+        pluginCacheAge = now;
+
+        auto insertIntoCache = [this](const QString &pluginPath) {
+            KPluginMetaData metadata(pluginPath);
+            if (!metadata.isValid()) {
+                qCWarning(LOG_PLASMA) << "invalid metadata" << pluginPath;
+                return;
+            }
+
+            plugins[metadata.pluginId()].append(metadata);
+        };
+
+        for (const QString &dir : dirs)
+            KPluginLoader::forEachPlugin(dir, insertIntoCache);
+    } else if (now - pluginCacheAge > maxCacheAge) {
+        // cache is old and we're not within a few seconds of startup anymore
+        useRuntimeCache = false;
+        plugins.clear();
+    }
+
+    //if name wasn't a path, pluginName == name
+    const QString pluginName = name.section(QLatin1Char('/'), -1);
+
+    QVector<KPluginMetaData> ret;
+
+    if (useRuntimeCache) {
+        auto it = plugins.constFind(pluginName);
+        if (it != plugins.constEnd()) {
+            ret = *it;
+        }
+        qCDebug(LOG_PLASMA) << "loading applet by name" << name << useRuntimeCache << ret.size();
+    } else {
+
+        for (const auto& dir : dirs) {
+            ret = KPluginLoader::findPluginsById(dir, pluginName);
+            if (!ret.isEmpty())
+                break;
+        }
+    }
+    return ret;
+}
+
+} // Plasma Namespace
